# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '5.5.0'
git_version = 'ce1d8837be043f3e5f66b96cfb0f59e9425ec598'
full_version = '5.5.0'
is_release = True

if not is_release:
    version = full_version